(* 
 * Copyright (c) 2002 The University of Wroclaw.
 * All rights reserved. See file COPYRIGHT for details about licensing.
 *
 * $Id: ciface.ml,v 1.1 2002/05/27 15:08:33 malekith Exp $
 *
 * Module implementation: Ciface
 * Description: C language interface.
 *)

open Gonttree
open Gontutil
open Gonttypes

let cize_name =
  let rx = Str.regexp ":" in
  function s -> Str.global_replace rx "_" s
  
let ksize_name =
  let rx = Str.regexp ":" in
  function s -> Str.global_replace rx "$" s

let strip_module =
  let rx = Str.regexp "^[^:]*::" in
  function s -> Str.global_replace rx "" s
  
let rec c_type = function
  | Int 
  | Bool -> "gont_ptroff"
  | Void -> "void"
  | Float -> "double*"
  | Func _ -> "struct gont_lambda_exp *"
  | String -> "struct gont_string *"
  | AbstractType
  | TyVar _ -> "void*"
  | Tuple _ -> "void**"
  | Struct _
  | OptStruct _ -> "void*"
  | Exn
  | Union _ -> "struct gont_union *"
  | (NamedType _) as t -> c_type (instance_value t)
  
let dump_header m =
  let dump_type td = 
    let dump_union_fld fr =
      let sym = match fr.fr_sym with Some x -> x | None -> assert false in
      let id = union_field_id sym in
      print_endline (xf "#define GontUF_%s %d" (cize_name sym.sym_name) id) in
    let dump_struct_fld fr =
      let sym = match fr.fr_sym with Some x -> x | None -> assert false in
      print_endline (xf "\t%s %s;" (c_type fr.fr_type) 
                                   (strip_module sym.sym_name)) in
    print_endline (xf "/* type %s */" td.td_name);
    match td.td_type with
    | Union flds -> List.iter dump_union_fld flds; print_newline ()
    | Struct flds | OptStruct flds -> 
        print_endline (xf "struct GontS_%s__%s {" td.td_module td.td_name);
        List.iter dump_struct_fld flds;
        print_endline "}\n"
    | _ -> () in
  let dump_global sym = 
    match sym.sym_kind with
    | VariableKind when (not (has_property sym IsUnionField)) ->
      begin
        match sym.sym_type with
        | Func {ft_ret_type = ret; ft_parms = prms} ->
          let prms' = String.concat ", " (List.map c_type prms) in
          print_endline (xf "%s %s(%s);" 
                            (c_type ret) 
                            (ksize_name sym.sym_name)
                            prms')
        | t ->
          print_endline (xf "%s %s;" (c_type t) (ksize_name sym.sym_name))
      end
    | _ -> ()
  in
  
  print_endline (xf "/* C header for Gont module: %s */" m);
  print_endline (xf "#ifndef GONT_MODULE_IFACE_%s_INCLUDED" m);
  print_endline (xf "#define GONT_MODULE_IFACE_%s_INCLUDED 1\n" m);
  let (g, t) = Gontmodule.globals m in
  List.iter dump_type (List.rev t);
  List.iter dump_global (List.rev g);
  print_endline (xf "#endif /* GONT_MODULE_IFACE_%s_INCLUDED */" m);
  0
;;
